"""Strategies for simple types like scalars or enums."""
from typing import Union

import graphql
from hypothesis import strategies as st

from ..types import ScalarValueNode

MIN_INT = -(2 ** 31)
MAX_INT = 2 ** 31 - 1


def scalar(type_: graphql.GraphQLScalarType, nullable: bool = True) -> st.SearchStrategy[ScalarValueNode]:
    if type_.name == "Int":
        return int_(nullable)
    if type_.name == "Float":
        return float_(nullable)
    if type_.name == "String":
        return string(nullable)
    if type_.name == "ID":
        return id_(nullable)
    if type_.name == "Boolean":
        return boolean(nullable)
    raise TypeError("Custom scalar types are not supported")


def enum(type_: graphql.GraphQLEnumType, nullable: bool = True) -> st.SearchStrategy[graphql.EnumValueNode]:
    enum_value = st.sampled_from(list(type_.values))
    return maybe_null(st.builds(graphql.EnumValueNode, value=enum_value), nullable)


def int_(nullable: bool = True) -> st.SearchStrategy[graphql.IntValueNode]:
    value = st.integers(min_value=MIN_INT, max_value=MAX_INT).map(str)
    return maybe_null(st.builds(graphql.IntValueNode, value=value), nullable)


def float_(nullable: bool = True) -> st.SearchStrategy[graphql.FloatValueNode]:
    value = st.floats(allow_infinity=False, allow_nan=False).map(str)
    return maybe_null(st.builds(graphql.FloatValueNode, value=value), nullable)


def string(nullable: bool = True) -> st.SearchStrategy[graphql.StringValueNode]:
    value = st.text(alphabet=st.characters(blacklist_categories=("Cs",), max_codepoint=0xFFFF))
    return maybe_null(st.builds(graphql.StringValueNode, value=value), nullable)


def id_(nullable: bool = True) -> st.SearchStrategy[Union[graphql.StringValueNode, graphql.IntValueNode]]:
    return string(nullable) | int_(nullable)


def boolean(nullable: bool = True) -> st.SearchStrategy[graphql.BooleanValueNode]:
    return maybe_null(st.builds(graphql.BooleanValueNode, value=st.booleans()), nullable)


def maybe_null(strategy: st.SearchStrategy, nullable: bool) -> st.SearchStrategy:
    if nullable:
        strategy |= st.just(graphql.NullValueNode())
    return strategy



"""Strategies are based on ANTLR4 grammar.

https://github.com/antlr/grammars-v4/blob/master/graphql/GraphQL.g4
"""
import string
from typing import Any, Optional, Tuple

from hypothesis import strategies as st


def schema(min_definitions: int = 1, max_definitions: int = 10) -> st.SearchStrategy[str]:
    # definition+;
    return st.lists(definition(), min_size=min_definitions, max_size=max_definitions).map("\n".join)


def definition() -> st.SearchStrategy[str]:
    # executableDefinition
    # 	| typeSystemDefinition
    # 	| typeSystemExtension;
    return st.one_of(
        typesystem_definition(),
    )


def typesystem_definition() -> st.SearchStrategy[str]:
    # schemaDefinition
    # 	| typeDefinition
    # 	| directiveDefinition
    # 	;
    return st.one_of(type_definition())


def type_definition() -> st.SearchStrategy[str]:
    # scalarTypeDefinition
    # 	| objectTypeDefinition
    # 	| interfaceTypeDefinition
    # 	| unionTypeDefinition
    # 	| enumTypeDefinition
    # 	| inputObjectTypeDefinition;
    return st.one_of(scalar_typedef(), object_typedef())


def scalar_typedef() -> st.SearchStrategy[str]:
    # description? 'scalar' name directives?

    def format_scalar(item: Tuple[Optional[str], str, Optional[str]]) -> str:
        return f"{item[0] or ''} scalar {item[1]} {item[2] or ''}".strip()

    return st.tuples(DESCRIPTION | st.none(), name(), directives() | st.none()).map(format_scalar)  # type: ignore


STRING = st.text(alphabet=string.ascii_letters).map('"{}"'.format)
BLOCK_STRING = st.text(alphabet=string.ascii_letters).map('"""{}"""'.format)
STRING_VALUE = STRING | BLOCK_STRING
DESCRIPTION = STRING_VALUE

FIRST_LETTER = string.ascii_letters + "_"
NON_FIRST_LETTER = FIRST_LETTER + string.digits


def name() -> st.SearchStrategy[str]:
    # [_A-Za-z] [_0-9A-Za-z]*
    return st.tuples(st.sampled_from(FIRST_LETTER), st.text(alphabet=NON_FIRST_LETTER)).map(lambda x: "{}{}".format(*x))


def directives() -> st.SearchStrategy[str]:
    # directive+;
    # Take only one directive, since in the core there are only "include" and "skip" which
    return directive()


def directive() -> st.SearchStrategy[str]:
    # '@' name arguments?;
    # TODO. restrict to "if: Boolean"

    def format_directive(item: Tuple[str, Optional[str]]) -> str:
        return f"@{item[0]} {item[1] or ''}".strip()

    return st.tuples(st.sampled_from(("include", "skip")), arguments() | st.none()).map(format_directive)  # type: ignore


def arguments() -> st.SearchStrategy[str]:
    # '(' argument+ ')';
    return st.lists(argument(), min_size=1).map(lambda x: "({})".format(", ".join(x)))


def argument() -> st.SearchStrategy[str]:
    # name ':' value;
    return st.tuples(name(), value()).map(lambda x: "{}:{}".format(*x))


def value() -> st.SearchStrategy[int]:
    # variable
    # 	| intValue
    # 	| floatValue
    # 	| stringValue
    # 	| booleanValue
    # 	| nullValue
    # 	| enumValue
    # 	| listValue
    # 	| objectValue
    #    ;
    return st.integers()


def object_typedef() -> st.SearchStrategy[str]:
    # description?   'type' name implementsInterfaces?  directives? fieldsDefinition?

    def format_object_typedef(item: Tuple[Optional[str], str, Optional[str], Optional[str], Optional[str]]) -> str:
        return f"{item[0] or ''} type {item[1]} {item[2] or ''} {item[3] or ''} {item[4] or ''}".strip().replace(
            "  ", " "
        )

    return st.tuples(
        DESCRIPTION | st.none(),
        name(),
        implements_interfaces() | st.none(),
        directives() | st.none(),
        fields_definition() | st.none(),
    ).map(
        format_object_typedef  # type: ignore
    )


@st.composite  # type: ignore
def implements_interfaces(draw: Any) -> Any:
    # 'implements' '&'? namedType | implementsInterfaces '&' namedType
    # TODO. namedType should be defined in the document
    def format_implements_interfaces(item: Tuple[Optional[str], str]) -> str:
        return f"implements {item[0] or ''} {item[1]}".replace("  ", " ")

    first = draw(st.tuples(st.just("&") | st.none(), name()).map(format_implements_interfaces))  # type: ignore
    types = " & ".join(draw(st.lists(name())))
    if types:
        types = f"& {types}"
    return f"{first} {types}".strip()


def fields_definition() -> st.SearchStrategy[str]:
    # '{' fieldDefinition+ '}';
    # TODO. names should be unique
    return st.lists(field_definition(), min_size=1).map(lambda item: "{{ {} }}".format(" ".join(item)))


def field_definition() -> st.SearchStrategy[str]:
    # description? name  argumentsDefinition? ':' type_  directives? ;

    def format_field_definition(item: Tuple[Optional[str], str, Optional[str], str, Optional[str]]) -> str:
        return f"{item[0] or ''} {item[1]} {item[2] or ''} : {item[3]} {item[4] or ''}".strip().replace("  ", " ")

    return st.tuples(
        DESCRIPTION | st.none(), name(), arguments_definition() | st.none(), TYPE, directives() | st.none()
    ).map(
        format_field_definition  # type: ignore
    )


def arguments_definition() -> st.SearchStrategy[str]:
    # '(' inputValueDefinition+ ')'
    return st.lists(input_value_definition(), min_size=1).map(lambda item: "( {} )".format(" ".join(item)))


def input_value_definition() -> st.SearchStrategy[str]:
    # description? name ':' type_ defaultValue? directives?;

    def format_input_value_definition(item: Tuple[Optional[str], str, str, Optional[str], Optional[str]]) -> str:
        return f"{item[0] or ''} {item[1]} : {item[2]} {item[3] or ''} {item[4] or ''}".strip().replace("  ", " ")

    # TODO. defaultvalue should match type
    return st.tuples(DESCRIPTION | st.none(), name(), TYPE, default_value() | st.none(), directives() | st.none(),).map(
        format_input_value_definition  # type: ignore
    )


def default_value() -> st.SearchStrategy[str]:
    # '=' value;
    return value().map("={}".format)


NAMED_TYPE = name()
# '[' type_ ']';
LIST_TYPE = st.deferred(lambda: TYPE.map("[{}]".format))
# namedType '!'?
#     | listType '!'?
#     ;
TYPE: st.SearchStrategy[str] = st.deferred(
    lambda: st.tuples(NAMED_TYPE, st.just("!") | st.none()).map(lambda item: f"{item[0]}{item[1] or ''}") | LIST_TYPE
)



"""Strategies for GraphQL queries."""
from functools import partial
from typing import Any, Callable, Dict, Generator, Iterable, List, Optional, Tuple, TypeVar, Union

import attr
import graphql
from hypothesis import strategies as st

from ..types import Field, InputTypeNode, InterfaceOrObject, SelectionNodes
from . import primitives


@attr.s(slots=True)
class Context:
    """The common context for query generation."""

    schema: graphql.GraphQLSchema = attr.ib()


def query(schema: Union[str, graphql.GraphQLSchema], fields: Optional[Iterable[str]] = None) -> st.SearchStrategy[str]:
    """A strategy for generating valid queries for the given GraphQL schema.

    The output query will contain a subset of fields defined in the `Query` type.

    :param schema: GraphQL schema as a string or `graphql.GraphQLSchema`.
    :param fields: Restrict generated fields to ones in this list.
    """
    if isinstance(schema, str):
        parsed_schema = graphql.build_schema(schema)
    else:
        parsed_schema = schema
    if parsed_schema.query_type is None:
        raise ValueError("Query type is not defined in the schema")
    if fields is not None:
        fields = tuple(fields)
        if not fields:
            raise ValueError("If you pass `fields`, it should not be empty")
        invalid_fields = tuple(field for field in fields if field not in parsed_schema.query_type.fields)
        if invalid_fields:
            raise ValueError(f"Unknown fields: {', '.join(invalid_fields)}")
    context = Context(parsed_schema)
    return _fields(context, parsed_schema.query_type, fields=fields).map(make_query).map(graphql.print_ast)


def _fields(
    context: Context,
    object_type: InterfaceOrObject,
    fields: Optional[Tuple[str, ...]] = None,
) -> st.SearchStrategy[SelectionNodes]:
    """Generate a subset of fields defined on the given type."""
    if fields:
        subset = {name: value for name, value in object_type.fields.items() if name in fields}
    else:
        subset = object_type.fields
    # minimum 1 field, an empty query is not valid
    return subset_of_fields(subset).flatmap(lambda f: lists_of_field_nodes(context, f))


make_selection_set_node = partial(graphql.SelectionSetNode, kind="selection_set")


def unwrap_field_type(field: Field) -> graphql.GraphQLNamedType:
    """Get the underlying field type which is not wrapped."""
    type_ = field.type
    while isinstance(type_, graphql.GraphQLWrappingType):
        type_ = type_.of_type
    return type_


def make_query(selections: SelectionNodes) -> graphql.DocumentNode:
    """Create top-level node for a query AST."""
    return graphql.DocumentNode(
        kind="document",
        definitions=[
            graphql.OperationDefinitionNode(
                kind="operation_definition",
                operation=graphql.OperationType.QUERY,
                selection_set=make_selection_set_node(selections=selections),
            )
        ],
    )


def field_nodes(context: Context, name: str, field: graphql.GraphQLField) -> st.SearchStrategy[graphql.FieldNode]:
    """Generate a single field node with optional children."""
    return st.builds(
        partial(graphql.FieldNode, name=graphql.NameNode(value=name)),
        arguments=list_of_arguments(context, field.args),
        selection_set=st.builds(make_selection_set_node, selections=fields_for_type(context, field)),
    )


def fields_for_type(
    context: Context,
    field: graphql.GraphQLField,
) -> st.SearchStrategy[Optional[SelectionNodes]]:
    """Extract proper type from the field and generate field nodes for this type."""
    field_type = unwrap_field_type(field)
    if isinstance(field_type, graphql.GraphQLObjectType):
        return _fields(context, field_type)
    if isinstance(field_type, graphql.GraphQLInterfaceType):
        # Besides the fields on the interface type, it is possible to generate inline fragments on types that
        # implement this interface type
        implementations = context.schema.get_implementations(field_type).objects
        if not implementations:
            # Shortcut when there are no implementations - take fields from the interface itself
            return _fields(context, field_type)
        variants: List[InterfaceOrObject] = [field_type, *implementations]
        return unique_by(variants, lambda v: v.name).flatmap(lambda t: interfaces(context, t))
    if isinstance(field_type, graphql.GraphQLUnionType):
        # A union is a set of object types - take a subset of them and generate inline fragments
        return unique_by(field_type.types, lambda m: m.name).flatmap(lambda m: inline_fragments(context, m))
    # Other types don't have fields
    return st.none()


def interfaces(context: Context, types: List[InterfaceOrObject]) -> st.SearchStrategy[SelectionNodes]:
    strategies = [
        inline_fragment(context, type_) if isinstance(type_, graphql.GraphQLObjectType) else _fields(context, type_)
        for type_ in types
    ]
    return st.tuples(*strategies).map(flatten).map(list)  # type: ignore


T = TypeVar("T")


def flatten(items: Tuple[Union[T, List[T]], ...]) -> Generator[T, None, None]:
    for item in items:
        if isinstance(item, list):
            yield from item
        else:
            yield item


def inline_fragments(context: Context, items: List[graphql.GraphQLObjectType]) -> st.SearchStrategy[SelectionNodes]:
    """Create inline fragment nodes for each given item."""
    return fixed_lists((inline_fragment(context, type_) for type_ in items))


def inline_fragment(
    context: Context, type_: graphql.GraphQLObjectType
) -> st.SearchStrategy[graphql.InlineFragmentNode]:
    """Build `InlineFragmentNode` for the given type."""
    return st.builds(
        partial(
            graphql.InlineFragmentNode, type_condition=graphql.NamedTypeNode(name=graphql.NameNode(value=type_.name))
        ),
        selection_set=st.builds(make_selection_set_node, selections=_fields(context, type_)),
    )


def list_of_arguments(
    context: Context, kwargs: Dict[str, graphql.GraphQLArgument]
) -> st.SearchStrategy[List[graphql.ArgumentNode]]:
    """Generate a list `graphql.ArgumentNode` for a field."""
    args = []
    for name, argument in kwargs.items():
        try:
            argument_strategy = argument_values(context, argument)
        except TypeError as exc:
            if not isinstance(argument.type, graphql.GraphQLNonNull):
                continue
            raise TypeError("Non-nullable custom scalar types are not supported as arguments") from exc
        args.append(
            st.builds(partial(graphql.ArgumentNode, name=graphql.NameNode(value=name)), value=argument_strategy)
        )
    return fixed_lists(args)


def fixed_lists(args: Iterable[st.SearchStrategy[T]]) -> st.SearchStrategy[List[T]]:
    return st.tuples(*args).map(list)


def unique_by(variants: List[T], key: Callable[[T], Any]) -> st.SearchStrategy[List[T]]:
    return st.lists(st.sampled_from(variants), min_size=1, unique_by=key)


def argument_values(context: Context, argument: graphql.GraphQLArgument) -> st.SearchStrategy[InputTypeNode]:
    """Value of `graphql.ArgumentNode`."""
    return value_nodes(context, argument.type)


def value_nodes(context: Context, type_: graphql.GraphQLInputType) -> st.SearchStrategy[InputTypeNode]:
    """Generate value nodes of a type, that corresponds to the input type.

    They correspond to all `GraphQLInputType` variants:
        - GraphQLScalarType -> ScalarValueNode
        - GraphQLEnumType -> EnumValueNode
        - GraphQLInputObjectType -> ObjectValueNode

    GraphQLWrappingType[T] is unwrapped:
        - GraphQLList -> ListValueNode[T]
        - GraphQLNonNull -> T (processed with nullable=False)
    """
    type_, nullable = check_nullable(type_)
    # Types without children
    if isinstance(type_, graphql.GraphQLScalarType):
        return primitives.scalar(type_, nullable)
    if isinstance(type_, graphql.GraphQLEnumType):
        return primitives.enum(type_, nullable)
    # Types with children
    if isinstance(type_, graphql.GraphQLList):
        return lists(context, type_, nullable)
    if isinstance(type_, graphql.GraphQLInputObjectType):
        return objects(context, type_, nullable)
    raise TypeError(f"Type {type_.__class__.__name__} is not supported.")


def check_nullable(type_: graphql.GraphQLInputType) -> Tuple[graphql.GraphQLInputType, bool]:
    """Get the wrapped type and detect if it is nullable."""
    nullable = True
    if isinstance(type_, graphql.GraphQLNonNull):
        type_ = type_.of_type
        nullable = False
    return type_, nullable


def lists(
    context: Context, type_: graphql.GraphQLList, nullable: bool = True
) -> st.SearchStrategy[graphql.ListValueNode]:
    """Generate a `graphql.ListValueNode`."""
    type_ = type_.of_type
    list_value = st.lists(value_nodes(context, type_))
    return primitives.maybe_null(st.builds(graphql.ListValueNode, values=list_value), nullable)


def objects(
    context: Context, type_: graphql.GraphQLInputObjectType, nullable: bool = True
) -> st.SearchStrategy[graphql.ObjectValueNode]:
    """Generate a `graphql.ObjectValueNode`."""
    fields_value = subset_of_fields(type_.fields, force_required=True).flatmap(
        lambda x: list_of_object_field_nodes(context, x)
    )
    return primitives.maybe_null(st.builds(graphql.ObjectValueNode, fields=fields_value), nullable)


def subset_of_fields(
    fields: Dict[str, Field], *, force_required: bool = False
) -> st.SearchStrategy[List[Tuple[str, Field]]]:
    """A helper to select a subset of fields."""
    field_pairs = sorted(fields.items())
    # if we need to always generate required fields, then return them and extend with a subset of optional fields
    if force_required:
        required, optional = [], {}
        for name, field in field_pairs:
            # TYPING: `field` is always `GraphQLInputField` as `force_required` equals `True` only with
            # `GraphQLInputObjectType`. A better solution is to create a separate function
            if graphql.is_required_input_field(field):  # type: ignore
                required.append((name, field))
            else:
                optional[name] = field
        if optional:
            return subset_of_fields(optional).map(lambda p: p + required)
        return st.just(required)
    # pairs are unique by field name
    return st.lists(st.sampled_from(field_pairs), min_size=1, unique_by=lambda x: x[0])


def object_field_nodes(
    context: Context, name: str, field: graphql.GraphQLInputField
) -> st.SearchStrategy[graphql.ObjectFieldNode]:
    return st.builds(
        partial(graphql.ObjectFieldNode, name=graphql.NameNode(value=name)),
        value=value_nodes(context, field.type),
    )


def list_of_nodes(
    context: Context,
    items: List[Tuple],
    strategy: Callable[[Context, str, Field], st.SearchStrategy],
) -> st.SearchStrategy[List]:
    return fixed_lists((strategy(context, name, field) for name, field in items))


list_of_object_field_nodes = partial(list_of_nodes, strategy=object_field_nodes)
lists_of_field_nodes = partial(list_of_nodes, strategy=field_nodes)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['hypothesis_graphql', 'hypothesis_graphql._strategies']

package_data = \
{'': ['*']}

install_requires = \
['attrs>20.3.0', 'graphql-core>=3.1.0,<4.0.0', 'hypothesis>5.8.0']

setup_kwargs = {
    'name': 'hypothesis-graphql',
    'version': '0.5.1',
    'description': 'Hypothesis strategies for GraphQL schemas and queries',
    'long_description': 'hypothesis-graphql\n==================\n\n|Build| |Coverage| |Version| |Python versions| |Chat| |License|\n\nHypothesis strategies for GraphQL schemas, queries and data.\n\n**NOTE** This package is experimental, some features are not supported yet.\n\nUsage\n-----\n\nThere are two strategies for different use cases.\n\n1. Schema generation - ``hypothesis_graphql.strategies.schema()``\n2. Query - ``hypothesis_graphql.strategies.query(schema)``.\n\nLets take this schema as an example:\n\n.. code::\n\n    type Book {\n      title: String\n      author: Author\n    }\n\n    type Author {\n      name: String\n      books: [Book]\n    }\n\n    type Query {\n      getBooks: [Book]\n      getAuthors: [Author]\n    }\n\nThen strategies might be used in this way:\n\n.. code:: python\n\n    from hypothesis import given\n    from hypothesis_graphql import strategies as gql_st\n\n    SCHEMA = "..."  # the one above\n\n\n    @given(query=gql_st.query(SCHEMA))\n    def test_query(query):\n        ...\n        # This query might be generated:\n        #\n        # query {\n        #   getBooks {\n        #     title\n        #   }\n        # }\n\nTo restrict the set of fields in generated queries use the ``fields`` argument:\n\n.. code:: python\n\n    ...\n\n\n    @given(query=gql_st.query(SCHEMA, fields=["getAuthors"]))\n    def test_query(query):\n        # Only `getAuthors` will be generated\n        ...\n\n.. |Build| image:: https://github.com/Stranger6667/hypothesis-graphql/workflows/build/badge.svg\n   :target: https://github.com/Stranger6667/hypothesis-graphql/actions\n.. |Coverage| image:: https://codecov.io/gh/Stranger6667/hypothesis-graphql/branch/master/graph/badge.svg\n   :target: https://codecov.io/gh/Stranger6667/hypothesis-graphql/branch/master\n   :alt: codecov.io status for master branch\n.. |Version| image:: https://img.shields.io/pypi/v/hypothesis-graphql.svg\n   :target: https://pypi.org/project/hypothesis-graphql/\n.. |Python versions| image:: https://img.shields.io/pypi/pyversions/hypothesis-graphql.svg\n   :target: https://pypi.org/project/hypothesis-graphql/\n.. |Chat| image:: https://img.shields.io/gitter/room/Stranger6667/hypothesis-graphql.svg\n   :target: https://gitter.im/Stranger6667/hypothesis-graphql\n   :alt: Gitter\n.. |License| image:: https://img.shields.io/pypi/l/hypothesis-graphql.svg\n   :target: https://opensource.org/licenses/MIT\n',
    'author': 'Dmitry Dygalo',
    'author_email': 'dadygalo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Stranger6667/hypothesis-graphql',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
